import matplotlib.pylab as plt
import numpy as np

from wildboar._utils import check_array_fast
from wildboar.datasets import load_dataset
from wildboar.distance import distance
from wildboar.distance.dtw import dtw_envelop, dtw_lb_keogh, dtw_pairwise_distance

check_array_fast(np.array([1, 2, 3, 2]).reshape(2, 2), ensure_2d=False, allow_nd=True)
x, y = load_dataset("GunPoint")

l, u = dtw_envelop(x[-1], r=10)
lu = dtw_lb_keogh(x[-1], x[0], r=10)
plt.plot(x[-1])
plt.plot(x[0])
plt.plot(l, "b--")
plt.plot(u, "r--")
plt.plot(lu, "g--")
plt.show()

x = np.array([[1, 10, 2, 3], [1, 2, 3, 4], [2, 1, 2, 2]])
#
print(
    distance(
        [1, 2, 3], x, metric="scaled_dtw", metric_params={"r": 0}, return_index=True
    )
)
print(dtw_pairwise_distance(x, r=1))



# cython: language_level=3

# This file is part of wildboar
#
# wildboar is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# wildboar is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser
# General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors: Isak Samsten
from libc.math cimport NAN
from libc.stdlib cimport free, malloc

from .._data cimport TSDatabase


cdef class FeatureEngineer:

    cdef Py_ssize_t init(self, TSDatabase *td) nogil:
        return 0

    cdef Py_ssize_t get_n_features(self, TSDatabase *td) nogil:
        return -1

    cdef Py_ssize_t get_n_outputs(self, TSDatabase *td) nogil:
        return self.get_n_features(td)

    cdef Py_ssize_t next_feature(
        self,
        Py_ssize_t feature_id,
        TSDatabase *td, 
        Py_ssize_t *samples, 
        Py_ssize_t n_samples,
        Feature *transient,
        size_t *seed,
    ) nogil:
        return -1

    cdef Py_ssize_t free_transient_feature(self, Feature *feature) nogil:
        return -1

    cdef Py_ssize_t free_persistent_feature(self, Feature *feature) nogil:
        return -1

    cdef Py_ssize_t init_persistent_feature(
        self, 
        TSDatabase *td,
        Feature *transient, 
        Feature *persistent
    ) nogil:
        return 0

    cdef double transient_feature_value(
        self,
        Feature *feature,
        TSDatabase *td,
        Py_ssize_t sample
    ) nogil:
        return NAN

    cdef double persistent_feature_value(
        self,
        Feature *feature,
        TSDatabase *td,
        Py_ssize_t sample
    ) nogil:
        return NAN

    cdef Py_ssize_t transient_feature_fill(
        self, 
        Feature *feature, 
        TSDatabase *td, 
        Py_ssize_t sample,
        TSDatabase *td_out,
        Py_ssize_t out_sample,
        Py_ssize_t out_feature,
    ) nogil:
        return -1

    cdef Py_ssize_t persistent_feature_fill(
        self, 
        Feature *feature, 
        TSDatabase *td, 
        Py_ssize_t sample,
        TSDatabase *td_out,
        Py_ssize_t out_sample,
        Py_ssize_t out_feature,
    ) nogil:
        return -1

    cdef object persistent_feature_to_object(self, Feature *feature):
        return None

    cdef Py_ssize_t persistent_feature_from_object(self, object object, Feature *feature):
        return 0

    cdef void transient_feature_values(
        self, 
        Feature *feature, 
        TSDatabase *td, 
        Py_ssize_t *samples, 
        Py_ssize_t n_samples,
        double* values
    ) nogil:
        cdef Py_ssize_t i
        for i in range(n_samples):
            values[i] = self.transient_feature_value(feature, td, samples[i])
    
    cdef void persistent_feature_values(
        self, 
        Feature *feature, 
        TSDatabase *td, 
        Py_ssize_t *samples, 
        Py_ssize_t n_samples,
        double* values
    ) nogil:
        cdef Py_ssize_t i
        for i in range(n_samples):
            values[i] = self.persistent_feature_value(feature, td, samples[i])


# This file is part of wildboar
#
# wildboar is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# wildboar is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser
# General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors: Isak Samsten

from ..embed import RocketEmbedding
from ._embed import EmbeddingRidgeClassifierCV, EmbeddingRidgeCV


class RocketClassifier(EmbeddingRidgeClassifierCV):
    def __init__(
        self,
        n_kernels=10000,
        *,
        kernel_size=None,
        sampling="auto",
        sampling_params=None,
        bias_prob=1.0,
        normalize_prob=1.0,
        padding_prob=0.5,
        alphas=(0.1, 1.0, 10.0),
        fit_intercept=True,
        normalize=False,
        scoring=None,
        cv=None,
        class_weight=None,
        n_jobs=None,
        random_state=None
    ):
        super().__init__(
            alphas=alphas,
            fit_intercept=fit_intercept,
            normalize=normalize,
            scoring=scoring,
            cv=cv,
            class_weight=class_weight,
            n_jobs=n_jobs,
            random_state=random_state,
        )
        self.padding_prob = padding_prob
        self.normalize_prob = normalize_prob
        self.bias_prob = bias_prob
        self.sampling = sampling
        self.sampling_params = sampling_params
        self.kernel_size = kernel_size
        self.n_kernels = n_kernels

    def _get_embedding(self, random_state):
        return RocketEmbedding(
            self.n_kernels,
            kernel_size=self.kernel_size,
            sampling=self.sampling,
            sampling_params=self.sampling_params,
            bias_prob=self.bias_prob,
            normalize_prob=self.normalize_prob,
            padding_prob=self.padding_prob,
            random_state=random_state,
            n_jobs=self.n_jobs,
        )


#
#
# class ShacketClassifier(EmbeddingRidgeClassifierCV):
#     def __init__(
#         self,
#         n_kernels=10000,
#         *,
#         size=None,
#         standardize=True,
#         use_bias=False,
#         alphas=(0.1, 1.0, 10.0),
#         fit_intercept=True,
#         normalize=False,
#         scoring=None,
#         cv=None,
#         class_weight=None,
#         n_jobs=None,
#         random_state=None
#     ):
#         super().__init__(
#             alphas=alphas,
#             fit_intercept=fit_intercept,
#             normalize=normalize,
#             scoring=scoring,
#             cv=cv,
#             class_weight=class_weight,
#             n_jobs=n_jobs,
#             random_state=random_state,
#         )
#         self.n_kernels = n_kernels
#         self.size = size
#         self.use_bias = use_bias
#         self.standardize = standardize
#
#     def _get_embedding(self, random_state):
#         return ShacketEmbedding(
#             n_kernels=self.n_kernels,
#             size=self.size,
#             standardize=self.standardize,
#             use_bias=self.use_bias,
#             n_jobs=self.n_jobs,
#             random_state=random_state,
#         )


class RocketRegressor(EmbeddingRidgeCV):
    def __init__(
        self,
        n_kernels=10000,
        *,
        alphas=(0.1, 1.0, 10.0),
        fit_intercept=True,
        normalize=False,
        scoring=None,
        cv=None,
        gcv_mode=None,
        n_jobs=None,
        random_state=None
    ):
        super().__init__(
            alphas=alphas,
            fit_intercept=fit_intercept,
            normalize=normalize,
            scoring=scoring,
            cv=cv,
            gcv_mode=gcv_mode,
            n_jobs=n_jobs,
            random_state=random_state,
        )
        self.n_kernels = n_kernels

    def _get_embedding(self, random_state):
        return RocketEmbedding(
            self.n_kernels, random_state=random_state, n_jobs=self.n_jobs
        )



# cython: language_level=3

# This file is part of wildboar
#
# wildboar is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# wildboar is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser
# General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

# Authors: Isak Samsten

cimport numpy as np

from ._utils import check_array_fast


def ts_database_dims(np.ndarray x):
    n_samples = x.shape[0]
    n_timestep = x.shape[x.ndim - 1]
    n_dims = 1
    if x.ndim == 3:
        n_dims = x.shape[1]
    return n_samples, n_dims, n_timestep

cdef TSDatabase ts_database_new(np.ndarray data):
    """Construct a new time series database from a ndarray """
    data = check_array_fast(data, allow_nd=True)
    if data.ndim < 2 or data.ndim > 3:
        raise ValueError("ndim {0} < 2 or {0} > 3".format(data.ndim))

    cdef TSDatabase sd
    sd.n_samples = <Py_ssize_t> data.shape[0]
    sd.n_timestep = <Py_ssize_t> data.shape[data.ndim - 1]
    sd.data = <double*> data.data
    sd.sample_stride = <Py_ssize_t> (data.strides[0] / <Py_ssize_t> data.itemsize)
    sd.timestep_stride = <Py_ssize_t> (data.strides[data.ndim - 1] / <Py_ssize_t> data.itemsize)

    if data.ndim == 3:
        sd.n_dims = <Py_ssize_t> data.shape[data.ndim - 2]
        sd.dim_stride = <Py_ssize_t> (data.strides[data.ndim - 2] / <Py_ssize_t> data.itemsize)
    else:
        sd.n_dims = 1
        sd.dim_stride = 0

    return sd


import numpy as np
from sklearn.model_selection import train_test_split

from wildboar.datasets import load_dataset
from wildboar.linear_model import RandomShapeletClassifier, RocketClassifier

x, y = load_dataset("ItalyPowerDemand", merge_train_test=True, preprocess="standardize")
x_train, x_test, y_train, y_test = train_test_split(x, y, random_state=123)
print(np.mean(x_train, axis=1))


f = RocketClassifier(
    n_kernels=10000,
    sampling="normal",
    sampling_params={"mean": 0, "scale": 1},
    n_jobs=16,
    random_state=123,
    alphas=np.logspace(-3, 3, 10),
    normalize=True,
)
f.fit(x_train, y_train)
print(f.score(x_test, y_test))


f = RandomShapeletClassifier(
    n_shapelets=10000,
    n_jobs=16,
    metric="scaled_euclidean",
    random_state=123,
    alphas=np.logspace(-3, 3, 10),
    normalize=True,
)
f.fit(x_train, y_train)
print(f.score(x_test, y_test))



# cython: cdivision=True
# cython: boundscheck=False
# cython: wraparound=False
# cython: language_level=3

# This file is part of wildboar
#
# wildboar is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# wildboar is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser
# General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

# Authors: Isak Samsten

cimport numpy as np

import numpy as np

from libc.math cimport INFINITY, floor, log2, pow, sqrt
from libc.stdlib cimport free, malloc
from libc.string cimport memcpy

from .._data cimport TSDatabase, ts_database_new
from .._utils cimport RAND_R_MAX, rand_int, rand_normal, rand_uniform
from ._feature cimport Feature, FeatureEngineer


cdef struct Rocket:
    Py_ssize_t length
    Py_ssize_t dilation
    Py_ssize_t padding
    bint return_mean
    double bias
    double *weight

cdef np.ndarray _to_ndarray(Py_ssize_t *arr, Py_ssize_t n):
    cdef Py_ssize_t i
    cdef np.ndarray out = np.zeros(n, dtype=int)
    for i in range(n):
        out[i] = arr[i]

    return out

cdef class WeightSampler:

    cdef void sample(
        self,
        TSDatabase *td,
        Py_ssize_t *samples,
        Py_ssize_t n_samples,
        double *weights,
        Py_ssize_t length,
        double *mean,
        size_t *seed
    ) nogil:
        pass

cdef class NormalWeightSampler(WeightSampler):
    cdef double mean
    cdef double scale

    def __init__(self, mean=0.0, scale=1.0):
        self.mean = mean
        self.scale = scale

    cdef void sample(
        self,
        TSDatabase *td,
        Py_ssize_t *samples,
        Py_ssize_t n_samples,
        double *weights,
        Py_ssize_t length,
        double *mean,
        size_t *seed
    ) nogil:
        cdef Py_ssize_t i
        mean[0] = 0
        for i in range(length):
            weights[i] = rand_normal(self.mean, self.scale, seed)
            mean[0] += weights[i]
        mean[0] = mean[0] / length

cdef class UniformWeightSampler(WeightSampler):

    cdef double lower
    cdef double upper

    def __init__(self, lower=-1.0, upper=1.0):
        self.lower = lower
        self.upper = upper

    cdef void sample(
        self,
        TSDatabase *td,
        Py_ssize_t *samples,
        Py_ssize_t n_samples,
        double *weights,
        Py_ssize_t length,
        double *mean,
        size_t *seed
    ) nogil:
        cdef Py_ssize_t i
        mean[0] = 0
        for i in range(length):
            weights[i] = rand_uniform(self.lower, self.upper, seed)
            mean[0] += weights[i]
        mean[0] = mean[0] / length

cdef class ShapeletWeightSampler(WeightSampler):
    cdef void sample(
        self,
        TSDatabase *td,
        Py_ssize_t *samples,
        Py_ssize_t n_samples,
        double *weights,
        Py_ssize_t length,
        double *mean,
        size_t *seed
    ) nogil:
        cdef Py_ssize_t start
        cdef Py_ssize_t index
        cdef Py_ssize_t dim
        cdef Py_ssize_t i, offset

        start = rand_int(0, td.n_timestep - length, seed)
        index = samples[rand_int(0, n_samples, seed)]
        if td.n_dims > 1:
            dim = rand_int(0, td.n_dims, seed)
        else:
            dim = 1

        offset = (
            index * td.sample_stride +
            start * td.timestep_stride +
            dim * td.dim_stride
        )
        mean[0] = 0
        for i in range(length):
            weights[i] = td.data[offset + i * td.timestep_stride]
            mean[0] += weights[i]
        mean[0] /= length


cdef class RocketFeatureEngineer(FeatureEngineer):
    cdef Py_ssize_t n_kernels
    cdef WeightSampler weight_sampler
    cdef double padding_prob
    cdef double bias_prob
    cdef double normalize_prob
    cdef Py_ssize_t *kernel_size
    cdef Py_ssize_t n_kernel_size

    def __cinit__(
        self,
        Py_ssize_t n_kernels,
        WeightSampler weight_sampler,
        np.ndarray kernel_size,
        double bias_prob,
        double padding_prob,
        double normalize_prob
    ):
        self.n_kernels = n_kernels
        self.weight_sampler = weight_sampler
        self.kernel_size = <Py_ssize_t*> malloc(sizeof(Py_ssize_t) * len(kernel_size))
        self.n_kernel_size = len(kernel_size)

        cdef Py_ssize_t i
        for i in range(len(kernel_size)):
            self.kernel_size[i] = kernel_size[i]

        self.bias_prob = bias_prob
        self.padding_prob = padding_prob
        self.normalize_prob = normalize_prob

    def __reduce__(self):
        return self.__class__, (
            self.n_kernels,
            self.weight_sampler,
            _to_ndarray(self.kernel_size, self.n_kernel_size),
            self.bias_prob,
            self.padding_prob,
            self.normalize_prob
        )

    def __dealloc__(self):
        free(self.kernel_size)

    cdef Py_ssize_t get_n_features(self, TSDatabase *td) nogil:
        return self.n_kernels

    cdef Py_ssize_t get_n_outputs(self, TSDatabase *td) nogil:
        return self.get_n_features(td) * 2

    cdef Py_ssize_t next_feature(
        self,
        Py_ssize_t feature_id,
        TSDatabase *td, 
        Py_ssize_t *samples, 
        Py_ssize_t n_samples,
        Feature *transient,
        size_t *seed
    ) nogil:
        cdef Rocket *rocket = <Rocket*> malloc(sizeof(Rocket))
        cdef Py_ssize_t i
        cdef double mean
        cdef Py_ssize_t length = self.kernel_size[rand_int(0, self.n_kernel_size, seed)]
        cdef double* weight = <double*> malloc(sizeof(double) * length)

        self.weight_sampler.sample(
            td, samples, n_samples, weight, length, &mean, seed
        )
        if rand_uniform(0, 1, seed) < self.normalize_prob:
            for i in range(length):
                weight[i] -= mean

        rocket.length = length
        rocket.dilation = <Py_ssize_t> floor(
            pow(2, rand_uniform(0, log2((td.n_timestep - 1) / (rocket.length - 1)), seed))
        )
        rocket.padding = 0
        if rand_uniform(0, 1, seed) < self.padding_prob:
            rocket.padding = ((rocket.length - 1) * rocket.dilation) // 2

        rocket.return_mean = rand_uniform(0, 1, seed) < 0.5
        rocket.weight = weight

        rocket.bias = 0
        if rand_uniform(0, 1, seed) < self.bias_prob:
            rocket.bias = rand_uniform(-1, 1, seed)

        if td.n_dims > 1:
            transient.dim = rand_int(0, td.n_dims, seed)
        else:
            transient.dim = 1

        transient.feature = rocket
        return 0

    cdef Py_ssize_t free_transient_feature(self, Feature *feature) nogil:
        cdef Rocket *rocket
        if feature.feature != NULL:
            rocket = <Rocket*> feature.feature
            if rocket.weight != NULL:
                free(rocket.weight)
            free(feature.feature)
        return 0

    cdef Py_ssize_t free_persistent_feature(self, Feature *feature) nogil:
        return self.free_transient_feature(feature)

    cdef Py_ssize_t init_persistent_feature(
        self, 
        TSDatabase *td,
        Feature *transient, 
        Feature *persistent
    ) nogil:
        cdef Rocket *transient_rocket = <Rocket*> transient.feature
        cdef Rocket *persistent_rocket = <Rocket*> malloc(sizeof(Rocket))
        
        persistent_rocket.dilation = transient_rocket.dilation
        persistent_rocket.bias = transient_rocket.bias
        persistent_rocket.length = transient_rocket.length
        persistent_rocket.padding = transient_rocket.padding
        persistent_rocket.return_mean = transient_rocket.return_mean
        persistent_rocket.weight = <double*> malloc(sizeof(double) * transient_rocket.length)
        memcpy(
            persistent_rocket.weight, 
            transient_rocket.weight, 
            sizeof(double) * transient_rocket.length,
        )
        persistent.dim = transient.dim
        persistent.feature = persistent_rocket
        return 0

    cdef double transient_feature_value(
        self,
        Feature *feature,
        TSDatabase *td,
        Py_ssize_t sample
    ) nogil:
        cdef double mean_val, max_val
        cdef Py_ssize_t sample_offset = (
            sample * td.sample_stride + feature.dim * td.dim_stride
        )
        cdef Rocket* rocket = <Rocket*> feature.feature
        apply_convolution(
            rocket.length,
            rocket.dilation,
            rocket.padding,
            rocket.bias,
            0,
            1,
            rocket.weight,
            sample_offset, 
            td.timestep_stride,
            td.n_timestep,
            td.data,
            &mean_val,
            &max_val,
        )
        if rocket.return_mean:
            return mean_val
        else:
            return max_val

    cdef double persistent_feature_value(
        self,
        Feature *feature,
        TSDatabase *td,
        Py_ssize_t sample
    ) nogil:
        return self.transient_feature_value(feature, td, sample)

    cdef Py_ssize_t transient_feature_fill(
        self, 
        Feature *feature, 
        TSDatabase *td, 
        Py_ssize_t sample,
        TSDatabase *td_out,
        Py_ssize_t out_sample,
        Py_ssize_t feature_id,
    ) nogil:
        cdef double mean_val, max_val
        cdef Py_ssize_t sample_offset = (
            sample * td.sample_stride + feature.dim * td.dim_stride
        )
        cdef Rocket* rocket = <Rocket*> feature.feature
        apply_convolution(
            rocket.length,
            rocket.dilation,
            rocket.padding,
            rocket.bias,
            0,
            1,
            rocket.weight,
            sample_offset,
            td.timestep_stride,
            td.n_timestep,
            td.data,
            &mean_val,
            &max_val,
        )
        cdef Py_ssize_t feature_offset = feature_id * 2 * td_out.timestep_stride
        cdef Py_ssize_t out_sample_offset = out_sample * td_out.sample_stride
        td_out.data[out_sample_offset + feature_offset] = mean_val
        td_out.data[out_sample_offset + feature_offset + td_out.timestep_stride] = max_val
        return 0

    cdef Py_ssize_t persistent_feature_fill(
        self, 
        Feature *feature, 
        TSDatabase *td, 
        Py_ssize_t sample,
        TSDatabase *td_out,
        Py_ssize_t out_sample,
        Py_ssize_t out_feature,
    ) nogil:
        return self.transient_feature_fill(
            feature, td, sample, td_out, out_sample, out_feature
        )

    cdef object persistent_feature_to_object(self, Feature *feature):
        cdef Py_ssize_t j
        cdef Rocket *rocket = <Rocket*> feature.feature

        weights = np.empty(rocket.length, dtype=np.float64)
        for j in range(rocket.length):
            weights[j] = rocket.weight[j]

        return feature.dim, (
            rocket.length,
            rocket.dilation,
            rocket.padding,
            weights,
            rocket.bias,
            rocket.return_mean
        )

    cdef Py_ssize_t persistent_feature_from_object(self, object object, Feature *feature):
        dim, (length, dilation, padding, weight, bias, return_mean) = object

        cdef Rocket *rocket = <Rocket*> malloc(sizeof(Rocket))
        rocket.length = length
        rocket.dilation = dilation
        rocket.padding = padding
        rocket.bias = bias
        rocket.weight = <double*> malloc(sizeof(double) * length)
        rocket.return_mean = return_mean
        
        cdef Py_ssize_t i
        for i in range(length):
            rocket.weight[i] = weight[i]

        feature.feature = rocket
        feature.dim = dim
        return 0

cdef void apply_convolution(
    Py_ssize_t length,
    Py_ssize_t dilation,
    Py_ssize_t padding,
    double bias,
    Py_ssize_t w_offset,
    Py_ssize_t w_stride,
    double *weight,
    Py_ssize_t x_offset,
    Py_ssize_t x_stride,
    Py_ssize_t x_length,
    double* x,
    double* mean_val,
    double* max_val
) nogil:
    cdef Py_ssize_t out_len, end
    cdef Py_ssize_t i, j, k
    cdef double ppv
    cdef double max
    cdef double sum

    out_len = (x_length + 2 * padding) - ((length - 1) * dilation)
    end = (x_length + padding) - ((length - 1) * dilation)
    max_val[0] = -INFINITY
    mean_val[0] = 0.0
    for i in range(-padding, end):
        inner_prod = bias
        k = i
        for j in range(length):
            if -1 < k < x_length:
                inner_prod += weight[w_offset + w_stride * j] * x[x_offset + x_stride * k]
            k += dilation
        if inner_prod > max_val[0]:
            max_val[0] = inner_prod

        if inner_prod > 0:
            mean_val[0] += 1

    mean_val[0] /= out_len


# cython: cdivision=True
# cython: boundscheck=False
# cython: wraparound=False
# cython: language_level=3

# This file is part of wildboar
#
# wildboar is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# wildboar is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser
# General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

# Authors: Isak Samsten

import numpy as np

cimport numpy as np
from libc.math cimport INFINITY, NAN, sqrt
from libc.stdlib cimport free, malloc, realloc

from .._data cimport TSDatabase, ts_database_new
from ._distance cimport DistanceMeasure, ScaledDistanceMeasure, TSCopy, TSView


cdef void simple_z_norm(TSCopy s, double *out) nogil:
    for i in range(s.length):
        out[i] = (s.data[0] - s.mean) / s.std

cdef int movestd(double *T,
                 Py_ssize_t offset,
                 Py_ssize_t stride,
                 Py_ssize_t length,
                 int m,
                 double *out) nogil:
    cdef double current_sum = 0
    cdef double current_sum_square = 0
    cdef double current_value
    cdef Py_ssize_t i
    cdef double *cumsum = <double*> malloc(sizeof(double) * length + 1)
    cdef double *cumsum_square = <double*> malloc(
        sizeof(double) * length + 1)

    cumsum[0] = 0
    cumsum_square[0] = 0
    for i in range(length):
        current_value = T[offset + i * stride]
        current_sum += current_value
        current_sum_square += current_value * current_value
        cumsum[i + 1] = current_sum
        cumsum_square[i + 1] = current_sum_square

    cdef double seq_sum = 0
    cdef double seq_sum_square = 0
    for i in range(length - m + 1):
        seq_sum = cumsum[m + i] - cumsum[i]
        seq_sum_square = cumsum_square[m + i] - cumsum_square[i]
        out[i] = sqrt(seq_sum_square / m - (seq_sum / m) ** 2)

cdef int simple_fft(Py_ssize_t offset,
                    Py_ssize_t stride,
                    Py_ssize_t length,
                    double *T,
                    Py_ssize_t out_offset,
                    Py_ssize_t out_stride,  # out_length = length - 1
                    complex *out) nogil:
    cdef int n = length
    cdef int rank = 1
    cdef fftw_plan p
    p = fftw_plan_many_dft_r2c(rank, &n, 1,
                               T + offset, NULL, stride, offset,
                               out + out_offset, NULL, out_stride, out_offset,
                               FFTW_ESTIMATE)

    fftw_execute(p)
    fftw_destroy_plan(p)
    fftw_cleanup()

cdef int simple_ifft(Py_ssize_t offset,
                     Py_ssize_t stride,
                     Py_ssize_t length,
                     complex *T,
                     Py_ssize_t out_offset,
                     Py_ssize_t out_stride,  # out_length = length + 1
                     double *out) nogil:
    cdef int n = length
    cdef int rank = 1
    cdef fftw_plan p
    p = fftw_plan_many_dft_c2r(rank, &n, 1,
                               T + offset, NULL, stride, offset,
                               out + out_offset, NULL, out_stride, out_offset,
                               FFTW_ESTIMATE)
    fftw_execute(p)

    cdef Py_ssize_t i
    for i in range(length):
        out[offset + i * out_stride] /= length

    fftw_destroy_plan(p)
    fftw_cleanup()

cpdef void test1(np.ndarray data, Py_ssize_t ts):
    cdef TSDatabase td = ts_database_new(data)
    cdef double *reconstruct = <double*> malloc(sizeof(double) * td.n_timestep)
    cdef complex *temp = <complex*> malloc(sizeof(complex) * td.n_timestep)

    cdef Py_ssize_t offset = ts * td.sample_stride + td.dim_stride
    print("offset", offset, td.n_timestep, td.timestep_stride, td.sample_stride)
    simple_fft(offset, td.timestep_stride, td.n_timestep, td.data,
               0, 1, temp)

    simple_ifft(0, 1, td.n_timestep, temp, 0, 1, reconstruct)

    cdef int i
    for i in range(td.n_timestep):
        print("reconstruct", reconstruct[i], temp[i], td.data[i])

cpdef void test() nogil:
    cdef int i
    cdef Py_ssize_t length = 10
    cdef double *T = <double*> malloc(sizeof(double) * length)
    for i in range(length):
        T[i] = i

    T[2] = 1
    T[4] = 2
    T[6] = 3
    T[8] = 10
    cdef double *out2 = <double*> malloc(sizeof(double) * length - 3 + 1)
    movestd(T, 0, 1, length, 3, out2)

    with gil:
        print("testing")
        for i in range(length):
            print("T", T[i])
        for i in range(2):
            print(out2[i])

    cdef complex *out3 = <complex*> malloc(sizeof(complex) * (4 - 1))
    simple_fft(2, 2, 4, T, 0, 1, out3)

    cdef double *out4 = <double*> malloc(sizeof(double) * 4)
    simple_ifft(0, 1, 4, out3, 0, 1, out4)
    with gil:
        for i in range(3):
            print("out3", out3[i])
        for i in range(4):
            print("out4", out4[i])

    cdef int n = 4
    cdef double *in1 = <double*> malloc(sizeof(double) * n)
    cdef complex *out = <complex*> malloc(sizeof(complex) * n)
    cdef double *in2 = <double*> malloc(sizeof(double) * n)

    cdef fftw_plan p, q

    p = fftw_plan_dft_r2c_1d(n, in1, out, FFTW_ESTIMATE)
    in1[0] = 1
    in1[1] = 2
    in1[2] = 3
    in1[3] = 10

    fftw_execute(p)

    q = fftw_plan_dft_c2r_1d(n, out, in2, FFTW_ESTIMATE);
    fftw_execute(q);

    with gil:
        for i in range(n):
            print(in1[i], out[i], in2[i] / 4)



# This file is part of wildboar
#
# wildboar is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# wildboar is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors: Isak Samsten

import os
import re

import numpy as np

from ._repository import (
    Bundle,
    JSONRepository,
    NpBundle,
    Repository,
    RepositoryCollection,
)
from .filter import make_filter
from .preprocess import named_preprocess

__all__ = [
    "Repository",
    "JSONRepository",
    "Bundle",
    "NpBundle",
    "set_cache_dir",
    "get_repository",
    "clear_cache",
    "list_repositories",
    "refresh_repositories",
    "list_bundles",
    "get_bundles",
    "install_repository",
    "load_dataset",
    "list_datasets",
    "load_datasets",
    "list_collections",
    "load_two_lead_ecg",
    "load_synthetic_control",
    "load_gun_point",
]

_CACHE_DIR = None
_REPOSITORIES = RepositoryCollection()


def _split_repo_bundle(repo_bundle_name):
    """Split a repository bundle string of the format {repo}/{bundle}

    Parameters
    ----------
    repo_bundle_name : str
        {repo}/{bundle} name

    Returns
    -------
    repository : str
        Key of the repository

    bundle : str
        Name of the bundle

    version : str
        An optional version

    tag : str
        An optional tag
    """
    match = re.match(
        r"(?:([a-zA-Z]+)/([a-zA-Z0-9\-]+))?(?::((?:\d+\.)?(?:\d+\.)?(?:\*|\d+)))?(?::([a-zA-Z\-]+))?$",  # noqa: E501
        repo_bundle_name,
    )
    if match:
        repository_name = match.group(1)
        bundle_name = match.group(2)
        version = match.group(3)
        if version:
            version = re.sub(r"(\d+\.\d+)((?:\.0+)*)$", "\\1", version)
        tag = match.group(4)
        return repository_name, bundle_name, version, tag
    else:
        raise ValueError("repository (%s) is not supported" % repo_bundle_name)


def set_cache_dir(cache_dir):
    """Change the global cache directory

    cache_dir : str
        The cache directory root
    """
    global _CACHE_DIR
    _CACHE_DIR = cache_dir


def _default_cache_dir():
    return _os_cache_path("wildboar") if not _CACHE_DIR else _CACHE_DIR


def _os_cache_path(dir):
    import platform

    if platform.system() == "Windows":
        cache_dir = os.path.expandvars(r"%LOCALAPPDATA%\cache")
        return os.path.join(cache_dir, dir)
    elif platform.system() == "Linux":
        cache_dir = os.environ.get("XDG_CACHE_HOME")
        if not cache_dir:
            cache_dir = ".cache"
        return os.path.join(os.path.expanduser("~"), cache_dir, dir)
    elif platform.system() == "Darwin":
        return os.path.join(os.path.expanduser("~"), "Library", "Caches", dir)
    else:
        return os.path.join(os.path.expanduser("~"), ".cache", dir)


def load_synthetic_control(merge_train_test=True):
    """Load the Synthetic_Control dataset

    See Also
    --------
    load_dataset : load a named dataset
    """
    return load_dataset(
        "SyntheticControl",
        repository="wildboar/ucr-tiny",
        merge_train_test=merge_train_test,
    )


def load_two_lead_ecg(merge_train_test=True):
    """Load the TwoLeadECG dataset

    See Also
    --------
    load_dataset : load a named dataset
    """
    return load_dataset(
        "TwoLeadECG",
        repository="wildboar/ucr-tiny",
        merge_train_test=merge_train_test,
    )


def load_gun_point(merge_train_test=True):
    """Load the GunPoint dataset

    See Also
    --------
    load_dataset : load a named dataset
    """
    return load_dataset(
        "GunPoint",
        repository="wildboar/ucr-tiny",
        merge_train_test=merge_train_test,
    )


def load_datasets(
    repository="wildboar/ucr",
    *,
    collection=None,
    cache_dir=None,
    create_cache_dir=True,
    progress=True,
    force=False,
    filter=None,
    **kwargs,
):
    """Load all datasets as a generator

    Parameters
    ----------
    repository : str
        The repository string

    collection : str, optional
        A collection of named datasets.

    progress : bool, optional
        If progress indicator is shown while downloading the repository.

    cache_dir : str, optional
        The cache directory for downloaded dataset repositories.

    create_cache_dir : bool, optional
        Create the cache directory if it does not exist.

    force : bool, optional
        Force re-download of cached repository

    filter: str, dict, list or callable, optional
        Filter the datasets

        - if callable, only yield those datasets for which the callable returns True.
          ``f(dataset, x, y) -> bool``

        - if dict, filter based on the keys and values, where keys are attributes and
          values comparison specs

        - if list, filter based on conjunction of attribute comparisons

        - if str, filter based on attribute comparison

        Attribute comparison
        --------------------

            The format of attribute comparisons are ``[attribute][comparison spec].

            Valid attributes are
            - ``dataset``
            - ``n_samples``
            - ``n_timestep``
            - ``n_dims``
            - ``n_labels``

        Comparison spec
        ---------------
            str of two parts, comparison operator (<, <=, >, >= or =) and a number,
            e.g., "<100", "<= 200", or ">300"

    kwargs : dict
        Optional arguments to ``load_dataset``

    Yields
    ------
    x : array-like
        Data samples

    y : array-like
        Data labels

    Examples
    --------

    >>> from wildboar.datasets import load_datasets
    >>> for dataset, (x, y) in load_datasets(repository='wildboar/ucr'):
    >>>     print(dataset, x.shape, y.shape)

    Print the names of datasets with more than 200 samples

    >>> for dataset, (x, y) in load_datasets(
    ...    repository='wildboar/ucr', filter={"n_samples": ">200"}
    ... ):
    >>>     print(dataset)

    >>> for dataset, (x, y) in load_datasets(
    ...    repository='wildboar/ucr', filter="n_samples>200"
    ... ):
    >>>     print(dataset)
    """
    for dataset in list_datasets(
        repository=repository,
        collection=collection,
        cache_dir=cache_dir,
        create_cache_dir=create_cache_dir,
        progress=progress,
        force=force,
    ):
        x, y = load_dataset(dataset, repository=repository, **kwargs)
        if filter is None:
            yield dataset, (x, y)
        elif hasattr(filter, "__call__"):
            if filter(dataset, x, y):
                yield dataset, (x, y)
        else:
            if make_filter(filter)(dataset, x, y):
                yield dataset, (x, y)


def load_dataset(
    name,
    *,
    repository="wildboar/ucr",
    dtype=None,
    preprocess=None,
    contiguous=True,
    merge_train_test=True,
    cache_dir=None,
    create_cache_dir=True,
    progress=True,
    return_extras=False,
    force=False,
    refresh=False,
    timeout=None,
):
    """Load a dataset from a repository

    Parameters
    ----------
    name : str
        The name of the dataset to load.

    repository : str, optional
        The data repository formatted as {repository}/{bundle}[:{version}][:{tag}]

    dtype : dtype, optional
        The data type of the returned data

    contiguous : bool, optional
        Ensure that the returned dataset is memory contiguous.

    preprocess : str or callable, optional
        Preprocess the dataset

        - if str, use named preprocess function (see ``preprocess._PREPROCESS.keys()``
          for valid keys)
        - if callable, function taking a np.ndarray and returns the preprocessed dataset

    merge_train_test : bool, optional
        Merge the existing training and testing partitions.

    progress: bool, optional
        Show a progress bar while downloading a bundle.

    cache_dir: str, optional
        The directory where downloaded files are cached

    create_cache_dir: bool, optional
        Create cache directory if missing (default=True)

    return_extras : bool, optional
        Return optional extras

        ..versionadded :: 1.1

    force : bool, optional
        Force re-download of already cached bundle

        ..versionadded :: 1.0.4

    refresh : bool, optional
        Refresh the repository

        ..versionadded :: 1.1

    timeout : float, optional
        Timeout for json request

        ..versionadded :: 1.1

    Returns
    -------
    x : ndarray
        The data samples, optional

    y : ndarray, optional
        The labels

    x_train : ndarray, optional
        The training samples if ``merge_train_test=False``

    x_test : ndarray, optional
        The testing samples if ``merge_train_test=False``

    y_train : ndarray, optional
        The training labels if ``merge_train_test=False``

    y_test : ndarray, optional
        The testing labels if ``merge_train_test=False``

    extras : dict, optional
        The optional extras if ``return_extras=True``

    Examples
    --------

    Load a dataset from the default repository

    >>> x, y = load_dataset("SyntheticControl")

    or if original training and testing splits are to be preserved

    >>> x_train, x_test, y_train, y_test = load_dataset(
    ...     "SyntheticControl", merge_train_test=False
    ... )

    or for a specific version of the dataset

    >>> x_train, x_test, y_train, y_test = load_dataset(
    ...     "Wafer", repository='wildboar/ucr-tiny:1.0'
    ... )
    """
    (
        repository_name,
        bundle_name,
        bundle_version,
        bundle_tag,
    ) = _split_repo_bundle(repository)
    dtype = dtype or np.float64
    cache_dir = cache_dir or _default_cache_dir()
    repository = get_repository(repository_name)

    if refresh:
        repository.refresh(timeout)

    ret_val = []
    x, y, n_train_samples, extras = repository.load_dataset(
        bundle_name,
        name,
        version=bundle_version,
        tag=bundle_tag,
        dtype=dtype,
        cache_dir=cache_dir,
        force=force,
        create_cache_dir=create_cache_dir,
        progress=progress,
    )
    if preprocess is None:

        def preprocess(identity):
            return identity

    elif isinstance(preprocess, str):
        preprocess = named_preprocess(preprocess)
    elif not hasattr(preprocess, "__call__"):
        raise ValueError("preprocess (%r) is not supported" % preprocess)

    if merge_train_test:
        ret_val.append(preprocess(x))
        ret_val.append(y)
    else:
        if n_train_samples == x.shape[0]:
            raise ValueError("found no test parts. Set merge_train_test=True.")

        ret_val.append(preprocess(x[:n_train_samples]))
        ret_val.append(preprocess(x[n_train_samples:]))
        if y is not None:
            ret_val.append(y[:n_train_samples])
            ret_val.append(y[n_train_samples:])
        else:
            ret_val.append(None)
            ret_val.append(None)
    if return_extras:
        ret_val.append(extras)

    if contiguous:
        return [
            np.ascontiguousarray(x)
            if isinstance(x, np.ndarray)
            else {k: np.ascontiguousarray(v) for k, v in extras.items()}
            for x in ret_val
        ]
    else:
        return ret_val


def list_datasets(
    repository="wildboar/ucr",
    *,
    collection=None,
    cache_dir=None,
    create_cache_dir=True,
    progress=True,
    force=False,
    refresh=False,
    timeout=None,
):
    """List the datasets in the repository

    Parameters
    ----------
    repository : str or Bundle, optional
        The data repository

        - if str load a named bundle, format {repository}/{bundle}

    collection : str, optional
        A collection of named datasets.

    progress: bool, optional
        Show a progress bar while downloading a bundle.

    cache_dir: str, optional
        The directory where downloaded files are cached (default='wildboar_cache')

    create_cache_dir: bool, optional
        Create cache directory if missing (default=True)

    force : bool, optional
        Force re-download of cached bundle

    refresh : bool, optional
        Refresh the repository

        ..versionadded :: 1.1

    timeout : float, optional
        Timeout for json request

        ..versionadded :: 1.1

    Returns
    -------
        dataset : set
            A set of dataset names
    """
    (
        repository_name,
        bundle_name,
        bundle_version,
        bundle_tag,
    ) = _split_repo_bundle(repository)
    cache_dir = cache_dir or _default_cache_dir()
    repository = get_repository(repository_name)
    if refresh:
        repository.refresh(timeout)
    return repository.list_datasets(
        bundle_name,
        collection=collection,
        version=bundle_version,
        tag=bundle_tag,
        cache_dir=cache_dir,
        create_cache_dir=create_cache_dir,
        progress=progress,
        force=force,
    )


def clear_cache(repository=None, *, cache_dir=None, keep_last_version=True):
    """Clear the cache by deleting cached datasets

    Parameters
    ----------
    repository : str, optional
        The name of the repository to clear cache.

        - if None, clear cache of all repositories

    cache_dir : str, optional
        The cache directory

    keep_last_version : bool, optional
        If true, keep the latest version of each repository.
    """
    cache_dir = cache_dir or _default_cache_dir()
    if repository is None:
        for repo in _REPOSITORIES:
            repo.clear_cache(cache_dir=cache_dir, keep_last_version=keep_last_version)
    else:
        get_repository(repository).clear_cache(
            cache_dir=cache_dir, keep_last_version=keep_last_version
        )


def get_repository(repository):
    """Get repository by name

    Parameters
    ----------
    repository : str
        Repository name

    Returns
    -------
    repository : Repository
        A repository
    """
    return _REPOSITORIES[repository]


def install_repository(repository, *, refresh=True, timeout=None):
    """Install repository

    Parameters
    ----------
    repository : str or Repository
        A repository

    refresh : bool, optional
        Refresh the repository

        ..versionadded :: 1.1

    timeout : float, optional
        Timeout for json request

        ..versionadded :: 1.1
    """
    if isinstance(repository, str):
        repository = JSONRepository(repository)
    _REPOSITORIES.append(repository, refresh=refresh, timeout=timeout)


def refresh_repositories(repository=None, *, timeout=None):
    """Refresh the installed repositories=None"""
    _REPOSITORIES.refresh(repository=repository, timeout=timeout)


def get_bundles(repository, *, refresh=False, timeout=None):
    """Get all bundles in the repository

    Parameters
    ----------
    repository : str
        Name of the repository

    refresh : bool, optional
        Refresh the repository

        ..versionadded :: 1.1

    timeout : float, optional
        Timeout for json request

        ..versionadded :: 1.1

    Returns
    -------
    dict : A dict of key Bundle pairs
    """
    repository = _REPOSITORIES[repository]
    if refresh:
        repository.refresh(timeout)

    return repository.get_bundles()


def list_bundles(repository, *, refresh=False, timeout=None):
    """Get a list of all bundle names in the specified repository.

    Parameters
    ----------
    repository : str
        The name of the repository

    refresh : bool, optional
        Refresh the repository

        ..versionadded :: 1.1

    timeout : float, optional
        Timeout for json request

        ..versionadded :: 1.1

    Returns
    -------
    bundle : str
        The name of the bundle
    """
    return [
        key
        for key, bundle in get_bundles(
            repository, refresh=refresh, timeout=timeout
        ).items()
    ]


def list_collections(repository):
    """List the collections of the repository

    Parameters
    ----------
    repository : str or Bundle, optional
        The data repository

        - if str load a named bundle, format {repository}/{bundle}

    Returns
    -------
    list : a list of collections
    """
    (
        repository_name,
        bundle_name,
        bundle_version,
        bundle_tag,
    ) = _split_repo_bundle(repository)
    repository = get_repository(repository_name)
    collections = repository.get_bundle(bundle_name).collections
    return list(collections.keys()) if collections is not None else []


def list_repositories(*, refresh=False, timeout=None):
    """List the key of all installed repositories

    refresh : bool, optional
        Refresh all repositories

        ..versionadded :: 1.1

    timeout : float, optional
        Timeout for json request

        ..versionadded :: 1.1
    """
    if refresh:
        refresh_repositories(timeout=timeout)
    return [repo.name for repo in _REPOSITORIES]


# Install the default 'wildboar' repository
def _get_dataset_version():
    from pkg_resources import parse_version

    from wildboar import __version__

    v = parse_version(__version__)
    if v.is_prerelease:
        return "master"
    else:
        return "%d.%d" % (v.major, v.minor)


install_repository(
    "https://isaksamsten.github.io/wildboar-datasets/%s/repo.json"
    % _get_dataset_version(),
    refresh=False,
)

refresh_repositories(timeout=1)



# This file is part of wildboar
#
# wildboar is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# wildboar is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser
# General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors: Isak Samsten

from ._shapelet_fast import RandomShapeletFeatureEngineer
from .base import BaseEmbedding


class RandomShapeletEmbedding(BaseEmbedding):
    """Embed a time series as the distances to a selection of random shapelets.

    Attributes
    ----------
    embedding_ : Embedding
        The underlying embedding object.

    References
    ----------
    Wistuba, Martin, Josif Grabocka, and Lars Schmidt-Thieme.
        Ultra-fast shapelets for time series classification.
        arXiv preprint arXiv:1503.05018 (2015).
    """

    def __init__(
        self,
        n_shapelets=1000,
        *,
        metric="euclidean",
        metric_params=None,
        min_shapelet_size=0,
        max_shapelet_size=1.0,
        n_jobs=None,
        random_state=None
    ):
        """
        Parameters
        ----------
        n_shapelets : int, optional
            The number of shapelets in the resulting embedding

        metric : str, optional
            The distance metric

            - if str use optimized implementations of the named distance measure
            - if callable a function taking two arrays as input

        metric_params: dict, optional
            Parameters to the metric

            - 'euclidean' and 'scaled_euclidean' take no parameters
            - 'dtw' and 'scaled_dtw' take a single paramater 'r'. If 'r' <= 1 it
            is interpreted as a fraction of the time series length. If > 1 it
            is interpreted as an exact time warping window. Use 'r' == 0 for
            a widow size of exactly 1.

        min_shapelet_size : float, optional
            Minimum shapelet size.

        max_shapelet_size : float, optional
            Maximum shapelet size.

        n_jobs : int, optional
            The number of jobs to run in parallel. None means 1 and
            -1 means using all processors.

        random_state : int or RandomState, optional
            The psudo-random number generator.
        """
        super().__init__(n_jobs=n_jobs)
        self.n_shapelets = n_shapelets
        self.metric = metric
        self.metric_params = metric_params
        self.min_shapelet_size = min_shapelet_size
        self.max_shapelet_size = max_shapelet_size
        self.random_state = random_state

    def _get_feature_engineer(self):
        if (
            self.min_shapelet_size < 0
            or self.min_shapelet_size > self.max_shapelet_size
        ):
            raise ValueError(
                "`min_shapelet_size` {0} <= 0 or {0} > {1}".format(
                    self.min_shapelet_size, self.max_shapelet_size
                )
            )
        if self.max_shapelet_size > 1:
            raise ValueError(
                "`max_shapelet_size` {0} > 1".format(self.max_shapelet_size)
            )
        max_shapelet_size = int(self.n_timestep_ * self.max_shapelet_size)
        min_shapelet_size = int(self.n_timestep_ * self.min_shapelet_size)
        if min_shapelet_size < 2:
            min_shapelet_size = 2
        return RandomShapeletFeatureEngineer(
            self.n_timestep_,
            self.metric,
            self.metric_params,
            min_shapelet_size,
            max_shapelet_size,
            self.n_shapelets,
        )



from sklearn.metrics import mean_squared_error
from sklearn.model_selection import train_test_split

from wildboar.datasets import load_dataset
from wildboar.ensemble import RockestRegressor, ShapeletForestRegressor
from wildboar.linear_model import RocketRegressor


def test(dataset, f, random_state=None):
    x, y = load_dataset(
        dataset,
        repository="wildboar/tsereg",
        merge_train_test=True,
        preprocess=None,
    )
    x_train, x_test, y_train, y_test = train_test_split(x, y, random_state=random_state)
    f.fit(x_train, y_train)
    print("%.6f" % mean_squared_error(y_test, f.predict(x_test), squared=False))


test(
    "FloodModeling1",
    ShapeletForestRegressor(
        metric="euclidean",
        n_estimators=100,
        n_jobs=-1,
        min_shapelet_size=0.9,
        max_shapelet_size=1.0,
        random_state=123,
    ),
    random_state=123,
)

test(
    "FloodModeling1",
    RocketRegressor(
        n_jobs=-1,
        normalize=True,
        random_state=123,
    ),
    random_state=123,
)

test(
    "FloodModeling1",
    RockestRegressor(
        n_kernels=100,
        n_jobs=-1,
        random_state=123,
    ),
    random_state=123,
)



# This file is part of wildboar
#
# wildboar is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# wildboar is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser
# General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors: Isak Samsten

from ._ensemble import (
    ExtraShapeletTreesClassifier,
    ExtraShapeletTreesRegressor,
    IsolationShapeletForest,
    RockestClassifier,
    RockestRegressor,
    ShapeletForestClassifier,
    ShapeletForestEmbedding,
    ShapeletForestRegressor,
)

__all__ = [
    "ShapeletForestClassifier",
    "ExtraShapeletTreesClassifier",
    "ShapeletForestRegressor",
    "ExtraShapeletTreesRegressor",
    "IsolationShapeletForest",
    "ShapeletForestEmbedding",
    "RockestRegressor",
    "RockestClassifier",
]



# This file is part of wildboar
#
# wildboar is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# wildboar is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser
# General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors: Isak Samsten

from libc.stdlib cimport free, malloc

from .._data cimport TSDatabase
from .._utils cimport RAND_R_MAX, rand_int
from ..distance._distance cimport (
    DistanceMeasure,
    TSCopy,
    TSView,
    get_distance_measure,
    ts_copy_free,
    ts_view_free,
)
from ._feature cimport Feature, FeatureEngineer


cdef class ShapeletFeatureEngineer(FeatureEngineer):

    cdef Py_ssize_t min_shapelet_size
    cdef Py_ssize_t max_shapelet_size

    cdef DistanceMeasure _distance_measure

    def __init__(
        self,
        n_timestep,
        object metric,
        dict metric_params,
        min_shapelet_size,
        max_shapelet_size,
    ):
        self._distance_measure = get_distance_measure(
            n_timestep, metric, metric_params
        )
        self.min_shapelet_size = min_shapelet_size
        self.max_shapelet_size = max_shapelet_size

    @property 
    def distance_measure(self):
        return self._distance_measure
    
    cdef Py_ssize_t init(self, TSDatabase *td) nogil:
        self._distance_measure.init(td)
        return 1

    cdef Py_ssize_t free_transient_feature(self, Feature *feature) nogil:
        if feature.feature != NULL:
            ts_view_free(<TSView*> feature.feature)
            free(feature.feature)

    cdef Py_ssize_t free_persistent_feature(self, Feature *feature) nogil:
        cdef TSCopy *ts_copy
        if feature.feature != NULL:
            ts_copy = <TSCopy*> feature.feature
            ts_copy_free(ts_copy)
            free(ts_copy)

    cdef Py_ssize_t init_persistent_feature(
        self, 
        TSDatabase *td,
        Feature *transient, 
        Feature *persistent
    ) nogil:
        cdef TSView *ts_view = <TSView*> transient.feature
        cdef TSCopy *ts_copy = <TSCopy*> malloc(sizeof(TSCopy))
        self._distance_measure.init_ts_copy(ts_copy, ts_view, td)
        persistent.dim = transient.dim
        persistent.feature = ts_copy
        return 1

    cdef double transient_feature_value(
        self,
        Feature *feature,
        TSDatabase *td,
        Py_ssize_t sample
    ) nogil:
        return self._distance_measure.ts_view_sub_distance(
            <TSView*> feature.feature, td, sample
        )

    cdef double persistent_feature_value(
        self,
        Feature *feature,
        TSDatabase *td,
        Py_ssize_t sample
    ) nogil:
        return self._distance_measure.ts_copy_sub_distance(
            <TSCopy*> feature.feature, td, sample
        )

    cdef Py_ssize_t transient_feature_fill(
        self, 
        Feature *feature, 
        TSDatabase *td, 
        Py_ssize_t sample,
        TSDatabase *td_out,
        Py_ssize_t out_sample,
        Py_ssize_t feature_id,
    ) nogil:
        cdef Py_ssize_t offset = (
            td_out.sample_stride * out_sample + 
            td_out.timestep_stride * feature_id
        )
        td_out.data[offset] = self._distance_measure.ts_view_sub_distance(
            <TSView*> feature.feature, td, sample
        )
        return 0

    cdef Py_ssize_t persistent_feature_fill(
        self, 
        Feature *feature, 
        TSDatabase *td, 
        Py_ssize_t sample,
        TSDatabase *td_out,
        Py_ssize_t out_sample,
        Py_ssize_t feature_id,
    ) nogil:
        cdef Py_ssize_t offset = (
            td_out.sample_stride * out_sample + 
            td_out.timestep_stride * feature_id
        )
        td_out.data[offset] = self._distance_measure.ts_copy_sub_distance(
            <TSCopy*> feature.feature, td, sample
        )
        return 0

    cdef object persistent_feature_to_object(self, Feature *feature):
        return feature.dim, self._distance_measure.object_from_ts_copy(<TSCopy*>feature.feature)

    cdef Py_ssize_t persistent_feature_from_object(self, object object, Feature *feature):
        cdef TSCopy *ts_copy = <TSCopy*> malloc(sizeof(TSCopy))
        dim, obj = object
        self._distance_measure.init_ts_copy_from_obj(ts_copy, obj)
        feature.dim = dim
        feature.feature = ts_copy
        return 1

cdef class RandomShapeletFeatureEngineer(ShapeletFeatureEngineer):

    cdef Py_ssize_t n_shapelets

    def __init__(
        self,
        n_timestep,
        object metric,
        dict metric_params,
        min_shapelet_size,
        max_shapelet_size,
        n_shapelets,
    ):
        super().__init__(
            n_timestep,
            metric, 
            metric_params,
            min_shapelet_size,
            max_shapelet_size,
        )
        self.n_shapelets = n_shapelets

    cdef Py_ssize_t get_n_features(self, TSDatabase *td) nogil:
        return self.n_shapelets

    cdef Py_ssize_t next_feature(
        self,
        Py_ssize_t feature_id,
        TSDatabase *td, 
        Py_ssize_t *samples, 
        Py_ssize_t n_samples,
        Feature *transient,
        size_t *random_seed
    ) nogil:
        if feature_id >= self.n_shapelets:
            return -1
        
        cdef Py_ssize_t shapelet_length
        cdef Py_ssize_t shapelet_start
        cdef Py_ssize_t shapelet_index
        cdef Py_ssize_t shapelet_dim
        cdef TSView *ts_view = <TSView*> malloc(sizeof(TSView))

        shapelet_length = rand_int(
            self.min_shapelet_size, self.max_shapelet_size, random_seed)
        shapelet_start = rand_int(
            0, td.n_timestep - shapelet_length, random_seed)
        shapelet_index = samples[rand_int(0, n_samples, random_seed)]
        if td.n_dims > 1:
            shapelet_dim = rand_int(0, td.n_dims, random_seed)
        else:
            shapelet_dim = 1

        transient.dim = shapelet_dim
        self._distance_measure.init_ts_view(
            td,
            ts_view,
            shapelet_index,
            shapelet_start,
            shapelet_length,
            shapelet_dim,
        )
        transient.feature = ts_view
        return 1


    


# cython: language_level=3

# This file is part of wildboar
#
# wildboar is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# wildboar is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser
# General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

# Authors: Isak Samsten

cimport numpy as np

from ..embed._feature cimport Feature, FeatureEngineer


# TODO: include impurity score...
cdef struct SplitPoint:
    Py_ssize_t split_point
    double threshold
    double impurity_improvement
    double impurity_left
    double impurity_right
    Feature feature

cdef class Tree:
    cdef FeatureEngineer feature_engineer

    cdef Py_ssize_t _max_depth
    cdef Py_ssize_t _capacity
    cdef Py_ssize_t _n_labels  # 1 for regression

    cdef Py_ssize_t _node_count
    cdef Py_ssize_t *_left
    cdef Py_ssize_t *_right
    cdef Feature **_features
    cdef double *_thresholds
    cdef double *_impurity
    cdef double *_values
    cdef double *_n_weighted_node_samples
    cdef Py_ssize_t *_n_node_samples

    cdef Py_ssize_t _increase_capacity(self) nogil except -1

    cdef Py_ssize_t add_leaf_node(
        self, 
        Py_ssize_t parent, 
        bint is_left, 
        Py_ssize_t n_node_samples, 
        double n_weighted_node_samples,
    ) nogil

    cdef void set_leaf_value(
        self,
        Py_ssize_t node_id,
        Py_ssize_t out_label,
        double out_value,
    ) nogil

    cdef Py_ssize_t add_branch_node(
        self, 
        Py_ssize_t parent, 
        bint is_left, 
        Py_ssize_t n_node_samples, 
        double n_weighted_node_samples,
        Feature *feature,
        double threshold,
        double impurity,
    ) nogil

    cpdef np.ndarray predict(self, object X)

    cpdef np.ndarray apply(self, object X)

    cpdef np.ndarray decision_path(self, object X)


# This file is part of wildboar
#
# wildboar is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# wildboar is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser
# General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors: Isak Samsten

from ._tree import (
    ExtraShapeletTreeClassifier,
    ExtraShapeletTreeRegressor,
    RocketTreeClassifier,
    RocketTreeRegressor,
    ShapeletTreeClassifier,
    ShapeletTreeRegressor,
)

__all__ = [
    "ShapeletTreeClassifier",
    "ShapeletTreeRegressor",
    "ExtraShapeletTreeClassifier",
    "ExtraShapeletTreeRegressor",
    "RocketTreeClassifier",
    "RocketTreeRegressor",
]



from sklearn.linear_model import RidgeClassifierCV
from sklearn.pipeline import make_pipeline

from wildboar.datasets import load_dataset
from wildboar.embed import RocketEmbedding

x_train, x_test, y_train, y_test = load_dataset("GunPoint", merge_train_test=False)


pipe = make_pipeline(RocketEmbedding(n_kernels=1000), RidgeClassifierCV())
pipe.fit(x_train, y_train)
print(pipe.score(x_test, y_test))



# This file is part of wildboar
#
# wildboar is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# wildboar is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser
# General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors: Isak Samsten
import numpy as np
from abc import ABCMeta, abstractmethod

from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.utils.validation import check_array, check_is_fitted, check_random_state

from ._embed_fast import (
    feature_embedding_fit,
    feature_embedding_fit_transform,
    feature_embedding_transform,
)

__all__ = [
    "BaseEmbedding",
]


class BaseEmbedding(TransformerMixin, BaseEstimator, metaclass=ABCMeta):
    """Base embedding

    Attributes
    ----------

    embedding_ : Embedding
        The underlying embedding.
    """

    def __init__(self, *, random_state=None, n_jobs=None):
        """
        Parameters
        ----------
        n_jobs : int, optional
            The number of jobs to run in parallel. None means 1 and
            -1 means using all processors.
        """
        self.n_jobs = n_jobs
        self.random_state = random_state

    def fit(self, x, y=None):
        """Fit the embedding.

        Parameters
        ----------
        x : array-like of shape [n_samples, n_timestep] or [n_samples, n_dimensions, n_timestep] # noqa: E501
            The time series dataset.

        y : None, optional
            For compatibility.

        Returns
        -------
        embedding : self
        """
        x = check_array(x)
        random_state = check_random_state(self.random_state)
        self.n_timestep_ = x.shape[-1]
        self.embedding_ = feature_embedding_fit(
            self._get_feature_engineer(), x, random_state, self.n_jobs
        )
        return self

    def transform(self, x):
        """Transform the dataset.

        Parameters
        ----------
        x : array-like of shape [n_samples, n_timestep] or [n_samples, n_dimensions, n_timestep] # noqa: E501
            The time series dataset.

        Returns
        -------
        x_embedding : ndarray of shape [n_samples, n_outputs]
            The embedding.
        """
        check_is_fitted(self, attributes="embedding_")
        x = check_array(x, dtype=np.float64, allow_nd=True, order="C")
        if x.ndim < 2 or x.ndim > 3:
            raise ValueError("illegal input dimensions")
        return feature_embedding_transform(self.embedding_, x, self.n_jobs)

    def fit_transform(self, x, y=None):
        """Fit the embedding and return the transform of x.

        Parameters
        ----------
        x : array-like of shape [n_samples, n_timestep] or [n_samples, n_dimensions, n_timestep] # noqa: E501
            The time series dataset.

        y : None, optional
            For compatibility.

        Returns
        -------
        x_embedding : ndarray of shape [n_samples, n_outputs]
            The embedding.
        """
        x = check_array(x, dtype=np.float64, allow_nd=True, order="C")
        if x.ndim < 2 or x.ndim > 3:
            raise ValueError("illegal input dimensions")

        random_state = check_random_state(self.random_state)
        self.n_timestep_ = x.shape[-1]
        embedding, x_out = feature_embedding_fit_transform(
            self._get_feature_engineer(), x, random_state, self.n_jobs
        )
        self.embedding_ = embedding
        return x_out

    @abstractmethod
    def _get_feature_engineer(self):
        pass



# cython: language_level=3

# This file is part of wildboar
#
# wildboar is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# wildboar is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser
# General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors: Isak Samsten
cimport numpy as np
from libc.stdlib cimport free, malloc

from .._data cimport TSDatabase, ts_database_new
from .._utils cimport RAND_R_MAX, safe_realloc
from ._feature cimport Feature, FeatureEngineer

from copy import deepcopy

import numpy as np
from joblib import Parallel, delayed, effective_n_jobs

from .._data import ts_database_dims


def clone_embedding(FeatureEngineer feature_engineer, features):
    cdef FeatureEmbedding embedding = FeatureEmbedding(feature_engineer, len(features))
    cdef Py_ssize_t i
    cdef Feature *feature
    for i in range(len(features)):
        feature = <Feature*> malloc(sizeof(Feature))
        feature_engineer.persistent_feature_from_object(features[i], feature)
        embedding.set_feature(i, feature)
    return embedding

def _partition_features(n_jobs, n_features, feature_engineer):
    n_jobs = min(effective_n_jobs(n_jobs), n_features)

    batch_size = n_features // n_jobs
    overflow = n_features % n_jobs
    feature_engineers = []
    feature_offsets = []
    current_offset = 0
    batch_sizes = []
    for i in range(n_jobs):
        feature_engineers.append(deepcopy(feature_engineer))
        current_overflow = 0
        if i < overflow:
            current_overflow = 1
        current_batch_size = batch_size + current_overflow
        feature_offsets.append(current_offset)
        batch_sizes.append(current_batch_size)
        current_offset += current_batch_size
    
    return n_jobs, feature_engineers, feature_offsets, batch_sizes


cdef class Batch:

    cdef list feature_engineers
    cdef FeatureEmbedding embedding
    cdef TSDatabase *x_out
    cdef TSDatabase *x_in

    def __init__(self, list feature_engineers, FeatureEmbedding embedding):
        self.feature_engineers = feature_engineers
        self.embedding = embedding

    cdef void init(self, TSDatabase *x_in, TSDatabase *x_out):
        self.x_in = x_in
        self.x_out = x_out

cdef class BatchTransform(Batch):

    def __call__(self, Py_ssize_t job_id, Py_ssize_t feature_offset, Py_ssize_t batch_size):
        cdef Py_ssize_t i, j
        cdef FeatureEngineer feature_engineer = self.feature_engineers[job_id]
        with nogil:
            for i in range(self.x_in.n_samples):
                for j in range(batch_size):
                    feature_engineer.persistent_feature_fill(
                        self.embedding.get_feature(feature_offset + j),
                        self.x_in,
                        i,
                        self.x_out,
                        i,
                        feature_offset + j,
                    )

cdef class BatchFitTransform(Batch):

    def __call__(self, Py_ssize_t job_id, Py_ssize_t feature_offset, Py_ssize_t batch_size, size_t seed):
        cdef Py_ssize_t i, j
        cdef FeatureEngineer feature_engineer = self.feature_engineers[job_id]
        cdef Py_ssize_t *samples = <Py_ssize_t*> malloc(sizeof(Py_ssize_t) * self.x_in.n_samples)
        cdef Feature transient_feature
        cdef Feature *persistent_feature

        with nogil:
            for i in range(self.x_in.n_samples):
                samples[i] = i

            for j in range(batch_size):
                feature_engineer.next_feature(
                    feature_offset + j,
                    self.x_in,
                    samples,
                    self.x_in.n_samples,
                    &transient_feature,
                    &seed,
                )
                persistent_feature = <Feature*> malloc(sizeof(Feature))
                feature_engineer.init_persistent_feature(
                    self.x_in, &transient_feature, persistent_feature
                )
                self.embedding.set_feature(feature_offset + j, persistent_feature)
                for i in range(self.x_in.n_samples):
                    feature_engineer.transient_feature_fill(
                        &transient_feature,
                        self.x_in,
                        i,
                        self.x_out,
                        i,
                        feature_offset + j,
                    )

                feature_engineer.free_transient_feature(&transient_feature)

cdef class FeatureEmbedding:

    cdef FeatureEngineer feature_engineer
    cdef Feature** _features
    cdef Py_ssize_t _n_features

    def __cinit__(self, FeatureEngineer feature_engineer, Py_ssize_t n_features):
        self.feature_engineer = feature_engineer
        self._features = <Feature**> malloc(sizeof(Feature*) * n_features)
        self._n_features = n_features

    def __dealloc__(self):
        cdef Py_ssize_t i
        for i in range(self._n_features):
            self.feature_engineer.free_persistent_feature(self._features[i])
            free(self._features[i])
        free(self._features)

    def __reduce__(self):
        return clone_embedding, (
            self.feature_engineer, 
            self.features,
        )

    cdef Py_ssize_t set_feature(self, Py_ssize_t i, Feature *feature) nogil:
        """Add a feature to the embedding

        feature : Feature*
            The feature.

            Note: The feature must be allocated using malloc
        """
        self._features[i] = feature
        return 0

    cdef Feature* get_feature(self, Py_ssize_t i) nogil:
        return self._features[i]

    @property
    def n_features(self):
        return self._n_features
    
    @property
    def features(self):
        return [
            self.feature_engineer.persistent_feature_to_object(self._features[i]) 
            for i in range(self._n_features)
        ]

    def __getitem__(self, item):
        if not isinstance(item, int) or 0 > item > self._n_features:
            raise ValueError()
        return self.feature_engineer.persistent_feature_to_object(self._features[item])

def feature_embedding_fit(FeatureEngineer feature_engineer, np.ndarray x, object random_state):
    cdef TSDatabase td = ts_database_new(x)
    cdef Py_ssize_t i
    cdef Feature transient_feature
    cdef Feature *persistent_feature
    cdef Py_ssize_t *samples = <Py_ssize_t*> malloc(sizeof(Py_ssize_t) * td.n_samples)
    cdef FeatureEmbedding embedding = FeatureEmbedding(
        feature_engineer, feature_engineer.get_n_features(&td),
    )
    cdef size_t seed = random_state.randint(0, RAND_R_MAX)
    
    with nogil:
        for i in range(td.n_samples):
            samples[i] = i
        
        for i in range(feature_engineer.get_n_features(&td)):
            persistent_feature = <Feature*> malloc(sizeof(Feature))
            feature_engineer.next_feature(
                i,
                &td,
                samples,
                td.n_samples,
                &transient_feature,
                &seed,
            )
            feature_engineer.init_persistent_feature(
                &td, &transient_feature, persistent_feature
            )
            feature_engineer.free_transient_feature(&transient_feature)
            embedding.set_feature(i, persistent_feature)

    return embedding

def feature_embedding_transform(FeatureEmbedding embedding, np.ndarray x, n_jobs=None):
    cdef TSDatabase x_in = ts_database_new(x)
    cdef FeatureEngineer feature_engineer = embedding.feature_engineer
    cdef Py_ssize_t n_outputs = feature_engineer.get_n_outputs(&x_in)
    cdef Py_ssize_t n_features = feature_engineer.get_n_features(&x_in)
    cdef np.ndarray out = np.empty((x.shape[0], n_outputs))
    cdef TSDatabase x_out = ts_database_new(out)

    n_jobs, feature_engineers, feature_offsets, batch_sizes = _partition_features(
        n_jobs, n_features, feature_engineer
    )
    cdef BatchTransform transform = BatchTransform(feature_engineers, embedding)
    transform.init(&x_in, &x_out)

    Parallel(n_jobs=n_jobs, prefer="threads")(
        delayed(transform)(
            jobid, feature_offsets[jobid], batch_sizes[jobid]
        )
        for jobid in range(n_jobs)
    )
    
    return out

def feature_embedding_fit_transform(FeatureEngineer feature_engineer, np.ndarray x, random_state, n_jobs=None):
    cdef TSDatabase x_in = ts_database_new(x)
    cdef Py_ssize_t n_outputs = feature_engineer.get_n_outputs(&x_in)
    cdef Py_ssize_t n_features = feature_engineer.get_n_features(&x_in)
    cdef FeatureEmbedding embedding = FeatureEmbedding(
        feature_engineer, n_features
    )
    cdef np.ndarray out = np.empty((x.shape[0], n_outputs))
    cdef TSDatabase x_out = ts_database_new(out)

    n_jobs, feature_engineers, feature_offsets, batch_sizes = _partition_features(
        n_jobs, n_features, feature_engineer
    )
    seeds = random_state.randint(0, RAND_R_MAX, size=n_jobs)

    cdef BatchFitTransform fit_transform = BatchFitTransform(feature_engineers, embedding)
    fit_transform.init(&x_in, &x_out)
    Parallel(n_jobs=n_jobs, prefer="threads")(
        delayed(fit_transform)(
            jobid, feature_offsets[jobid], batch_sizes[jobid], seeds[jobid]
        )
        for jobid in range(n_jobs)
    )

    return embedding, out


# This file is part of wildboar
#
# wildboar is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# wildboar is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser
# General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors: Isak Samsten
from ..embed import RandomShapeletEmbedding
from ._embed import EmbeddingRidgeClassifierCV, EmbeddingRidgeCV


class RandomShapeletClassifier(EmbeddingRidgeClassifierCV):
    def __init__(
        self,
        n_shapelets=1000,
        *,
        metric="euclidean",
        metric_params=None,
        min_shapelet_size=0.1,
        max_shapelet_size=1.0,
        alphas=(0.1, 1.0, 10.0),
        fit_intercept=True,
        normalize=False,
        scoring=None,
        cv=None,
        class_weight=None,
        n_jobs=None,
        random_state=None
    ):
        super().__init__(
            alphas=alphas,
            fit_intercept=fit_intercept,
            normalize=normalize,
            scoring=scoring,
            cv=cv,
            class_weight=class_weight,
            n_jobs=n_jobs,
            random_state=random_state,
        )
        self.n_shapelets = n_shapelets
        self.metric = metric
        self.metric_params = metric_params
        self.min_shapelet_size = min_shapelet_size
        self.max_shapelet_size = max_shapelet_size

    def _get_embedding(self, random_state):
        return RandomShapeletEmbedding(
            self.n_shapelets,
            metric=self.metric,
            metric_params=self.metric_params,
            min_shapelet_size=self.min_shapelet_size,
            max_shapelet_size=self.max_shapelet_size,
            random_state=random_state,
            n_jobs=self.n_jobs,
        )


class RandomShapeletRegressor(EmbeddingRidgeCV):
    def __init__(
        self,
        n_shapelets=1000,
        *,
        metric="euclidean",
        metric_params=None,
        min_shapelet_size=0.1,
        max_shapelet_size=1.0,
        alphas=(0.1, 1.0, 10.0),
        fit_intercept=True,
        normalize=False,
        scoring=None,
        cv=None,
        gcv_mode=None,
        n_jobs=None,
        random_state=None
    ):
        super().__init__(
            alphas=alphas,
            fit_intercept=fit_intercept,
            normalize=normalize,
            scoring=scoring,
            cv=cv,
            gcv_mode=gcv_mode,
            n_jobs=n_jobs,
            random_state=random_state,
        )
        self.n_shapelets = n_shapelets
        self.metric = metric
        self.metric_params = metric_params
        self.min_shapelet_size = min_shapelet_size
        self.max_shapelet_size = max_shapelet_size

    def _get_embedding(self, random_state):
        return RandomShapeletEmbedding(
            self.n_shapelets,
            metric=self.metric,
            metric_params=self.metric_params,
            min_shapelet_size=self.min_shapelet_size,
            max_shapelet_size=self.max_shapelet_size,
            random_state=random_state,
            n_jobs=self.n_jobs,
        )



# This file is part of wildboar
#
# wildboar is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# wildboar is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser
# General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors: Isak Samsten
import numpy as np


def named_preprocess(name):
    if name in _PREPROCESS:
        return _PREPROCESS[name]
    else:
        raise ValueError("preprocess (%s) does not exists" % name)


def standardize(x):
    """Scale x along the time dimension to have zero mean and unit standard deviation

    Parameters
    ----------
    x : ndarray of shape (n_samples, n_timestep) or (n_samples, n_dims, n_timestep)
        The dataset

    Returns
    -------
    x : ndarray of shape (n_samples, n_timestep) or (n_samples, n_dims, n_timestep)
        The standardized dataset
    """
    return (x - np.nanmean(x, axis=-1, keepdims=True)) / np.nanstd(
        x, axis=-1, keepdims=True
    )


normalize = standardize
normalize.__doc__ = standardize.__doc__


def minmax_scale(x, min=0, max=1):
    """Scale x along the time dimension so that each value is between min and max

    Parameters
    ----------
    x : ndarray of shape (n_samples, n_timestep) or (n_samples, n_dims, n_timestep)
        The dataset

    min : float, optional
        The minimum value

    max : float, optional
        The maximum value

    Returns
    -------
    x : ndarray of shape (n_samples, n_timestep) or (n_samples, n_dims, n_timestep)
        The transformed dataset
    """
    if min > max:
        raise ValueError("min must be smaller than max.")
    x_min = np.nanmin(x, axis=-1, keepdims=True)
    x_max = np.nanmax(x, axis=-1, keepdims=True)
    x = (x - x_min) / (x_max - x_min)
    return x * (max - min) + min


def maxabs_scale(x):
    """Scale each time series by its maximum absolute value.

    Parameters
    ----------
    x : ndarray of shape (n_samples, n_timestep) or (n_samples, n_dims, n_timestep)
        The dataset

    Returns
    -------
    x : ndarray of shape (n_samples, n_timestep) or (n_samples, n_dims, n_timestep)
        The transformed dataset
    """
    x_max = np.nanmax(np.abs(x), axis=-1, keepdims=True)
    return x / x_max


_PREPROCESS = {
    "standardize": standardize,
    "normalize": normalize,
    "minmax_scale": minmax_scale,
    "maxabs_scale": maxabs_scale,
}



# cython: cdivision=True
# cython: boundscheck=False
# cython: wraparound=False

# This file is part of wildboar
#
# wildboar is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# wildboar is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser
# General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

# Authors: Isak Samsten

#ctypedef double complex[2]

cdef extern from 'fftw3.h':
    ctypedef struct fftw_plan_struct:
        pass

    ctypedef fftw_plan_struct *fftw_plan

    fftw_plan fftw_plan_r2r_1d(int n,
                               double *in_,
                               double *out_,
                               int kind,
                               unsigned b) nogil

    fftw_plan fftw_plan_dft_r2c_1d(int n,
                                   double *in_,
                                   complex *out_,
                                   unsigned flags) nogil

    fftw_plan fftw_plan_dft_c2r_1d(int n,
                                   complex *in_,
                                   double *out_,
                                   unsigned flags) nogil

    fftw_plan fftw_plan_many_dft_r2c(int rank, const int *n, int howmany,
                                     double *in_, const int *inembed,
                                     int istride, int idist,
                                     complex *out, const int *onembed,
                                     int ostride, int odist,
                                     unsigned flags) nogil

    fftw_plan fftw_plan_many_dft_c2r(int rank, const int *n, int howmany,
                                     complex *in_, const int *inembed,
                                     int istride, int idist,
                                     double *out, const int *onembed,
                                     int ostride, int odist,
                                     unsigned flags) nogil

    void fftw_execute(fftw_plan plan) nogil

    void fftw_destroy_plan(fftw_plan plan) nogil

    void fftw_cleanup() nogil

    cdef int FFTW_ESTIMATE

cdef enum:
    FFTW_FORWARD = -1
    FFTW_BACKWARD = 1



# This file is part of wildboar
#
# wildboar is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# wildboar is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser
# General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

# Authors: Isak Samsten

from abc import ABCMeta, abstractmethod

import numpy as np
from sklearn.base import BaseEstimator, ClassifierMixin, RegressorMixin
from sklearn.linear_model import RidgeClassifierCV, RidgeCV
from sklearn.pipeline import Pipeline
from sklearn.utils.validation import check_array, check_is_fitted, check_random_state


class BaseEmbeddingEstimator(BaseEstimator, metaclass=ABCMeta):
    def __init__(self, *, random_state=None, n_jobs=None):
        self.random_state = random_state
        self.n_jobs = n_jobs

    def fit(self, x, y, sample_weight=None):
        x = check_array(x, dtype=np.float64, allow_nd=True, order="C")
        if x.ndim < 2 or x.ndim > 3:
            raise ValueError("illegal input dimensions")

        y = check_array(y, ensure_2d=False, order="C")
        random_state = check_random_state(self.random_state)
        self.pipe_ = Pipeline(
            [
                ("embedding", self._get_embedding(random_state.randint(2 ** 31))),
                ("estimator", self._get_estimator(random_state.randint(2 ** 31))),
            ],
        )
        self.pipe_.fit(x, y, estimator__sample_weight=sample_weight)
        return self

    @abstractmethod
    def _get_embedding(self, random_state):
        pass

    @abstractmethod
    def _get_estimator(self, random_state):
        pass


class EmbeddingClassifierMixin:
    def predict(self, x):
        check_is_fitted(self)
        x = check_array(x, dtype=np.float64, allow_nd=True, order="C")
        if x.ndim < 2 or x.ndim > 3:
            raise ValueError("illegal input dimensions")
        return self.pipe_.predict(x)

    def predict_proba(self, x):
        check_is_fitted(self)
        x = check_array(x, dtype=np.float64, allow_nd=True, order="C")
        if x.ndim < 2 or x.ndim > 3:
            raise ValueError("illegal input dimensions")
        return self.pipe_.predict_proba(x)

    def predict_log_proba(self, x):
        check_is_fitted(self)
        x = check_array(x, dtype=np.float64, allow_nd=True, order="C")
        if x.ndim < 2 or x.ndim > 3:
            raise ValueError("illegal input dimensions")
        return self.pipe_.predict_log_proba(x)

    def decision_function(self, x):
        check_is_fitted(self)
        x = check_array(x, dtype=np.float64, allow_nd=True, order="C")
        if x.ndim < 2 or x.ndim > 3:
            raise ValueError("illegal input dimensions")
        return self.pipe_.decision_function(x)


class EmbeddingRegressorMixin:
    def predict(self, x):
        check_is_fitted(self)
        x = check_array(x, dtype=np.float64, allow_nd=True, order="C")
        if x.ndim < 2 or x.ndim > 3:
            raise ValueError("illegal input dimensions")
        return self.pipe_.predict(x)

    def decision_function(self, x):
        check_is_fitted(self)
        x = check_array(x, dtype=np.float64, allow_nd=True, order="C")
        if x.ndim < 2 or x.ndim > 3:
            raise ValueError("illegal input dimensions")
        return self.pipe_.decision_function(x)


class EmbeddingRidgeClassifierCV(
    ClassifierMixin, EmbeddingClassifierMixin, BaseEmbeddingEstimator
):
    def __init__(
        self,
        *,
        alphas=(0.1, 1.0, 10.0),
        fit_intercept=True,
        normalize=False,
        scoring=None,
        cv=None,
        class_weight=None,
        n_jobs=None,
        random_state=None,
    ):
        super().__init__(random_state=random_state, n_jobs=n_jobs)
        self.alphas = alphas
        self.fit_intercept = fit_intercept
        self.normalize = normalize
        self.scoring = scoring
        self.cv = cv
        self.class_weight = class_weight
        self.random_state = random_state

    def _get_estimator(self, random_state):
        return RidgeClassifierCV(
            alphas=self.alphas,
            fit_intercept=self.fit_intercept,
            normalize=self.normalize,
            scoring=self.scoring,
            cv=self.cv,
            class_weight=self.class_weight,
            store_cv_values=False,
        )


class EmbeddingRidgeCV(RegressorMixin, EmbeddingRegressorMixin, BaseEmbeddingEstimator):
    def __init__(
        self,
        *,
        alphas=(0.1, 1.0, 10.0),
        fit_intercept=True,
        normalize=False,
        scoring=None,
        cv=None,
        gcv_mode=None,
        n_jobs=None,
        random_state=None,
    ):
        super().__init__(random_state=random_state, n_jobs=n_jobs)
        self.alphas = np.asarray(alphas)
        self.fit_intercept = fit_intercept
        self.normalize = normalize
        self.scoring = scoring
        self.cv = cv
        self.gcv_mode = gcv_mode

    def _get_estimator(self, random_state):
        return RidgeCV(
            alphas=self.alphas,
            fit_intercept=self.fit_intercept,
            normalize=self.normalize,
            scoring=self.scoring,
            cv=self.cv,
            gcv_mode=self.gcv_mode,
            store_cv_values=False,
            alpha_per_target=False,
        )



# This file is part of wildboar
#
# wildboar is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# wildboar is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser
# General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors: Isak Samsten
import numpy as np

from ._rocket_fast import (
    NormalWeightSampler,
    RocketFeatureEngineer,
    ShapeletWeightSampler,
    UniformWeightSampler,
)
from .base import BaseEmbedding

_SAMPLING_METHOD = {
    "auto": NormalWeightSampler,
    "normal": NormalWeightSampler,
    "uniform": UniformWeightSampler,
    "shapelet": ShapeletWeightSampler,
}


class RocketEmbedding(BaseEmbedding):
    """Embedd a time series using random convolution features

    Attributes
    ----------
    embedding_ : Embedding
        The underlying embedding

    References
    ----------
    Dempster, Angus, François Petitjean, and Geoffrey I. Webb.
        ROCKET: exceptionally fast and accurate time series classification using
        random convolutional kernels.
        Data Mining and Knowledge Discovery 34.5 (2020): 1454-1495.
    """

    def __init__(
        self,
        n_kernels=1000,
        *,
        sampling="auto",
        sampling_params=None,
        kernel_size=None,
        bias_prob=1.0,
        normalize_prob=1.0,
        padding_prob=0.5,
        n_jobs=None,
        random_state=None
    ):
        """
        Parameters
        ----------
        n_kernels : int, optional
            The number of kernels.

        n_jobs : int, optional
            The number of jobs to run in parallel. None means 1 and
            -1 means using all processors.

        random_state : int or RandomState, optional
            The psuodo-random number generator.
        """
        super().__init__(n_jobs=n_jobs)
        self.sampling = sampling
        self.sampling_params = sampling_params
        self.kernel_size = kernel_size
        self.bias_prob = bias_prob
        self.normalize_prob = normalize_prob
        self.padding_prob = padding_prob
        self.n_kernels = n_kernels
        self.random_state = random_state

    def _get_feature_engineer(self):
        if self.kernel_size is None:
            kernel_size = [7, 11, 13]
        elif isinstance(self.kernel_size, tuple) and len(self.kernel_size) == 2:
            min_size, max_size = self.kernel_size
            if min_size < 0 or min_size > max_size:
                raise ValueError(
                    "`min_size` {0} <= 0 or {0} > {1}".format(min_size, max_size)
                )
            if max_size > 1:
                raise ValueError("`max_size` {0} > 1".format(max_size))
            max_size = int(self.n_timestep_ * max_size)
            min_size = int(self.n_timestep_ * min_size)
            if min_size < 2:
                min_size = 2
            kernel_size = np.arange(min_size, max_size)
        else:
            kernel_size = self.kernel_size

        if self.sampling in _SAMPLING_METHOD:
            sampling_params = (
                {} if self.sampling_params is None else self.sampling_params
            )
            weight_sampler = _SAMPLING_METHOD[self.sampling](**sampling_params)
        else:
            raise ValueError("sampling (%r) is not supported." % self.sampling)
        return RocketFeatureEngineer(
            int(self.n_kernels),
            weight_sampler,
            np.array(kernel_size, dtype=int),
            float(self.bias_prob),
            float(self.padding_prob),
            float(self.normalize_prob),
        )


#
# class ShacketEmbedding(BaseEmbedding):
#     def __init__(
#         self,
#         n_kernels=1000,
#         *,
#         size=None,
#         standardize=True,
#         use_bias=False,
#         n_jobs=None,
#         random_state=None
#     ):
#         super().__init__(random_state=random_state, n_jobs=n_jobs)
#         self.n_kernels = n_kernels
#         self.standardize = standardize
#         self.use_bias = use_bias
#         self.size = size
#
#     def _get_feature_engineer(self):
#         if self.size is None:
#             size = [7, 11, 13]
#         elif isinstance(self.size, tuple) and len(self.size) == 2:
#             min_size, max_size = self.size
#             if min_size < 0 or min_size > max_size:
#                 raise ValueError(
#                     "`min_size` {0} <= 0 or {0} > {1}".format(min_size, max_size)
#                 )
#             if max_size > 1:
#                 raise ValueError("`max_size` {0} > 1".format(max_size))
#             max_size = int(self.n_timestep_ * max_size)
#             min_size = int(self.n_timestep_ * min_size)
#             if min_size < 2:
#                 min_size = 2
#             size = np.arange(min_size, max_size)
#         else:
#             size = self.size
#         return ShacketFeatureEngineer(
#             int(self.n_kernels),
#             np.array(size, dtype=int),
#             bool(self.use_bias),
#             bool(self.standardize),
#         )
